//! One way to ensure safe concurrency is message passing.
//! Threads send data between one another.
//! To accomplish message-sending concurrency, Rust
//! provides an implementation of channels. Channel
//! is generally a way for data to sent from one
//! channel to another.
//!
//! A channel has two halves: a transmitter and a reciever.
//! Channel is *closed* if either the transmitter or
//! reciever is dropped.

use std::sync::mpsc;
use std::thread;
use std::time::Duration;

fn main() {
    // mpsc stands for multiple producer, single consumer.
    // Rust can have multilple sending and only one receiving
    // tx for transmitter and rx for receiver.
    let (tx, rx) = mpsc::channel();

    // spawned thread needs to own transmitter to send messages.
    thread::spawn(move || {
        let val = String::from("hi");
        tx.send(val).unwrap();
        // Can't use after `send` call because the receiving
        // thread may modify the value.
        // println!("val is {val}");
    });

    // receiver has two useful methods: `recv` and `try_recv`.
    // `recv` will block main thread's execution.
    // `try_recv` useful if this thread has other work to do
    // while waiting for messages.
    let received = rx.recv().unwrap();
    println!("Got: {received}");

    // Sending multiple values and seeing the reciever waiting

    println!("Sending multiple values and seeing the receiver waiting section");

    let (tx, rx) = mpsc::channel();

    thread::spawn(move || {
        let vals = vec![
            String::from("hi"),
            String::from("from"),
            String::from("the"),
            String::from("thread"),
        ];

        for val in vals {
            tx.send(val).unwrap();
            thread::sleep(Duration::from_secs(1));
        }
    });

    for received in rx {
        println!("Got: {received}");
    }

    // Creating multiple producers by cloning the transmitter

    println!("Creating multiple producers by cloning the transmitter");

    let (tx, rx) = mpsc::channel();

    let tx1 = tx.clone();
    thread::spawn(move || {
        let vals = vec![
            String::from("hi"),
            String::from("from"),
            String::from("the"),
            String::from("thread"),
        ];

        for val in vals {
            tx1.send(val).unwrap();
            thread::sleep(Duration::from_secs(1));
        }
    });

    thread::spawn(move || {
        let vals = vec![
            String::from("more"),
            String::from("messages"),
            String::from("for"),
            String::from("you"),
        ];

        for val in vals {
            tx.send(val).unwrap();
            thread::sleep(Duration::from_secs(1));
        }
    });

    for received in rx {
        println!("Got: {received}");
    }
}
